@@ -12,7 +12,7 @@
 #include "response.h"
 #include "log.h"
 #include "client.h"
-#include "socket.h"
+#include "util.h"
 
 #define ACCEPT_TIMEOUT_SECS 1
 #define SHORT_TIMEOUT_SECS 2
@@ -215,7 +215,8 @@ static inline int
 process_wsgi_app(client_t *cli)
 {
     PyObject *args = NULL, *start = NULL, *res = NULL;
-    
+    PyGreenlet *greenlet;
+    ClientObject *pyclient;
     start = create_start_response(cli);
 
     if(!start){
@@ -224,11 +225,22 @@ process_wsgi_app(client_t *cli)
     args = Py_BuildValue("(OO)", cli->environ, start);
 
     current_client = PyDict_GetItemString(cli->environ, "meinheld.client");
+    pyclient = (ClientObject *)current_client;
+
 #ifdef DEBUG
     printf("start environ %p \n", cli->environ);
 #endif
-    res = PyObject_CallObject(wsgi_app, args);
+
+    //new greenlet
+    greenlet = PyGreenlet_New(wsgi_app, NULL);
+    // set_greenlet
+    pyclient->greenlet = greenlet;
+    Py_INCREF(pyclient->greenlet);
+
+    res = PyGreenlet_Switch(greenlet, args, NULL);
+    //res = PyObject_CallObject(wsgi_app, args);
     Py_DECREF(args);
+    Py_DECREF(greenlet);
     
 
     //check response & PyErr_Occurred
@@ -677,7 +689,6 @@ setup_server_env(void)
     setup_static_env(server_name, server_port);
     setup_start_response();
     setup_client();
-    setup_nsocket();
     PycString_IMPORT;
     PyGreenlet_Import();
     hub_switch_value = Py_BuildValue("(i)", -1);
@@ -1120,11 +1131,15 @@ meinheld_suspend_client(PyObject *self, PyObject *args)
     ClientObject *pyclient;
     client_t *client;
     PyGreenlet *parent;
-    int timeout;
+    int timeout = 0;
 
     if (!PyArg_ParseTuple(args, "O|i:_suspend_client", &temp, &timeout)){
         return NULL;
     }
+    if(timeout < 0){
+        PyErr_SetString(PyExc_ValueError, "timeout value out of range ");
+        return NULL;
+    }
     
     // check client object
     if(!CheckClientObject(temp)){
@@ -1226,17 +1241,81 @@ meinheld_resume_client(PyObject *self, PyObject *args)
 }
 
 PyObject *
-meinheld_get_socket_fromfd(PyObject *self, PyObject *args)
+meinheld_cancel_wait(PyObject *self, PyObject *args)
 {
     int fd;
-    if (!PyArg_ParseTuple(args, "i:_get_socket_fromfd", &fd)){
+    if (!PyArg_ParseTuple(args, "i:cancel_event", &fd)){
         return NULL;
     }
-    if(!((ClientObject *)current_client)->greenlet){
-        PyErr_SetString(PyExc_ValueError, "greenlet is not set");
+
+    if(fd < 0){
+        PyErr_SetString(PyExc_ValueError, "fileno value out of range ");
         return NULL;
     }
-    return NSocketObject_New(fd, (ClientObject *)current_client);
+    picoev_del(main_loop, fd);
+    Py_RETURN_NONE;
+}
+
+static inline void
+trampolin_switch_callback(picoev_loop* loop, int fd, int events, void* cb_arg)
+{
+    ClientObject *pyclient = (ClientObject *)cb_arg;
+
+    if ((events & PICOEV_TIMEOUT) != 0) {
+        PyErr_SetString(PyExc_IOError, "timeout");
+        switch_wsgi_app(loop, fd, (PyObject *)pyclient);
+    } else if ((events & PICOEV_WRITE) != 0 ||  (events & PICOEV_READ) != 0) {
+        switch_wsgi_app(loop, fd, (PyObject *)pyclient);
+    }
+}
+
+static inline PyObject*
+meinheld_trampolin(PyObject *self, PyObject *args, PyObject *kwargs)
+{
+    PyGreenlet *current, *parent;
+    ClientObject *pyclient;
+    int fd, event, timeout = 0;
+    PyObject *read = Py_None, *write = Py_None;
+
+	static char *keywords[] = {"fileno", "read", "write", "timeout", NULL};
+	
+    if (!PyArg_ParseTupleAndKeywords(args, kwargs, "i|OOi:trampolin", keywords, &fd, &read, &write, &timeout)){
+		return NULL;
+    }
+    
+    if(fd < 0){
+        PyErr_SetString(PyExc_ValueError, "fileno value out of range ");
+        return NULL;
+    }
+    
+    if(timeout < 0){
+        PyErr_SetString(PyExc_ValueError, "timeout value out of range ");
+        return NULL;
+    }
+
+    if(PyObject_IsTrue(read) && PyObject_IsTrue(write)){
+        event = PICOEV_READWRITE;
+    }else if(PyObject_IsTrue(read)){
+        event = PICOEV_READ;
+    }else if(PyObject_IsTrue(write)){
+        event = PICOEV_WRITE;
+    }else{
+        event = PICOEV_TIMEOUT;
+        if(timeout <= 0){
+            PyErr_SetString(PyExc_ValueError, "timeout value out of range ");
+            return NULL;
+        }
+    }
+
+    pyclient =(ClientObject *) current_client;
+    
+    picoev_del(main_loop, fd);
+    picoev_add(main_loop, fd, event, timeout, trampolin_switch_callback, (void *)pyclient);
+   
+    // switch to hub
+    current = pyclient->greenlet;
+    parent = PyGreenlet_GET_PARENT(current);
+    return PyGreenlet_Switch(parent, hub_switch_value, NULL);
 
 }
 
@@ -1268,7 +1347,9 @@ static PyMethodDef WsMethods[] = {
     {"_suspend_client", meinheld_suspend_client, METH_VARARGS, "resume client"},
     {"_resume_client", meinheld_resume_client, METH_VARARGS, "resume client"},
     // io
-    {"_get_socket_fromfd", meinheld_get_socket_fromfd, METH_VARARGS, "get socket fromfd"},
+    {"cancel_wait", meinheld_cancel_wait, METH_VARARGS, "cancel wait"},
+    {"trampolin", meinheld_trampolin, METH_VARARGS | METH_KEYWORDS, "trampolin"},
+
     {NULL, NULL, 0, NULL}        /* Sentinel */
 };
 
@@ -1289,11 +1370,7 @@ initserver(void)
         return;
     }
 
-    if(PyType_Ready(&NSocketObjectType) < 0){
-        return;
-    }
-
-	timeout_error = PyErr_NewException("meinheld.server.timeout",
+    timeout_error = PyErr_NewException("meinheld.server.timeout",
 					  PyExc_IOError, NULL);
 	if (timeout_error == NULL)
 		return;
